/**********************************************************************
    Path Sampling Simulation of Maier Stein Nonequilibrium System.

    For background on this system, and further references see
        "Irreversibility of classical fluctuations studied in 
            analogue electrical circuits"
    D.G. Luchinsky, P.V.E. McClintock, Nature v389 p463 (1997)
    
    The integrator is based on exaple code donated by Dmitrii G Luchinsky.

    See the comments in duffingPS.c for a description of the general idea.
    This system is different in that the system is nonequilibrium because of 
    a nongradient force field, rather than a time dependant force. 
    
    Usage:   lmmms D alpha mu samples  >outFile
    Example: lmmms 0.05 6.67 1.5  8192 >out

    
    8/98-2/99 Gavin Crooks  gavinc@garnet.berkeley.edu
                http://gold.cchem.berkeley.edu/~gavinc
        
        
**********************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdarg.h>

/* Random Number Generators.
 *  genRanf() returns a double [0,1)
 *  genRanGaussian returns a double distributed as a normal distribution 
 *      of unit varience.
 */
#include "genRan.h"                     

int     relaxPath(int relax);
int     initPath();
void    propagate(int start, int end);
int     propagateTest(int start);

void    printTimeAxis();
void    printPaths(int samples, int trans);
void    clearTriedFailed();
void    printTriedFailed();

void    die(char *comment, ...);


#define false   (0)
#define true    (1)

#define TIME        (16)                    /*Total time */
#define TSTEPSPER   (512)                   /*Time steps per time unit */
#define TSTEP       (1./TSTEPSPER)          /*time per time step*/
#define TBINS       (TIME*TSTEPSPER)        /*Total number of values that must be stored*/
#define TTEST       (TBINS/2)               /*Time for initial path drag*/
#define TPRINT      (8)                     /*controles How much path to print*/
#define SHOOTS      (TBINS/2)               /*Approximate number of shoots per independent sample*/
#define SKIP        (TSTEPSPER/16)          /*Output x every SKIP time*/


/* System paremeters */
double  D;                      /*Variance of fluctuating force*/
double  alpha;
double  mu; 
int     SAMPLES;                /*Number of samples */

/* These variables specify the history of the system */
double  x[TBINS];               /* x at time t  */
double  y[TBINS];               /* y at time t  */

double  grx[TBINS];             /* Random force at time t   */
double  gry[TBINS];             /* Random force at time t   */

int     tried[TBINS]= {0};      /*keep track of trials and failures*/
int     failed[TBINS]= {0};


/*********************************************************/
main(  int argc , char *argv[])
{   

    int i;
    int samples,t,total;
    
    int trans;          /*time when transition first occures*/
    
    /***********************************
            Initilisation
    ***********************************/
    if(argc!=5) die("%s D alpha mu samples ",argv[0]);

    sscanf(argv[1],"%lf",&D);
    sscanf(argv[2],"%lf",&alpha);
    sscanf(argv[3],"%lf",&mu);
    sscanf(argv[4],"%d",&SAMPLES);

    
    printf("LMMMS PS\t D=%lg\t alpha=%lg\t mu=%lg\t SAMPLES=%d\n"
                ,D,alpha,mu,SAMPLES);
        
        
    printTimeAxis();
    
    

    /************************************************
        Generate a sampling of transition paths. 
    *************************************************/

    /*** Generate an Initial Path, and print number of initialisation steps ***/
    printf("Init Steps %d\n",initPath(0.0));
    fflush(stdout);

    /* Relax to "equilibrium" path */
    relaxPath(SHOOTS*SAMPLES/16);
        
        
    clearTriedFailed();
    for(samples=0;samples<SAMPLES;samples++)
    {   
        trans = relaxPath(SHOOTS);
        printPaths(samples,trans);
    }

    printTriedFailed();
    
    
} 


/*************  Path relaxation Routines **********************************/

/*************************************************
    Generate a sequence of new paths, obeying the
    boundary conditions x[t]<0.0 for some t<TBINS  
*************************************************/
int relaxPath(int relax)
{
    int loop;
    int i;
    
    int tChanged;
    
    double oldgrX;          /*Storage for old gr. Needed if shoot fails */
    double oldgrY;          /*Storage for old gr. Needed if shoot fails */
    
    int trans;
    
    trans=propagateTest(0);
    
    /*** Insanity Check ***/
    if(trans >= TBINS ) die("Error:RelaxPath: trans:%d !",trans);
    
    for(loop=0;loop<relax;loop++)
    {
        /*Attempt to change path at one point*/
        tChanged = (int) floor(genRanf()*TBINS);  /* choose gr[] to change*/

        oldgrX = grx[tChanged];
        oldgrY = gry[tChanged];
        grx[tChanged] =sqrt(D*TSTEP)*genRanGaussian(); /*Generate a gr with correct variance*/
        gry[tChanged] =sqrt(D*TSTEP)*genRanGaussian(); /*Generate a gr with correct variance*/

        /*if tChanged>trans then new path is automatically accepted */
        if(tChanged<=trans) 
        {
            tried[trans-tChanged]++;
            trans=propagateTest(tChanged);
        
            if(trans==TBINS)
            {   /*Failed*/
                grx[tChanged]=oldgrX;
                gry[tChanged]=oldgrY;
                trans = propagateTest(tChanged);        
                failed[trans-tChanged]++;
            }
        }
    
        /* Ocasionally try a sign flip move.
         *  If sucessful then a single simulation
         *  Should generate both up and down paths
         */
        if(  genRanf()<0.1)
        {
            tChanged = (int) floor(genRanf()*TBINS);  /* choose gr[] to change*/
            for(i=tChanged;i<TBINS;i++) gry[i] =-gry[i];    
            
            /*if tChanged>trans then new path is automatically accepted */
            if(tChanged<=trans) 
            {
                trans=propagateTest(tChanged);
        
                if(trans==TBINS)
                {   /*Failed*/
                    for(i=tChanged;i<TBINS;i++) gry[i] =-gry[i];
                    trans = propagateTest(tChanged);        
                }
            }
        }   
    }
    
    propagate(0,TBINS-1);
    return(trans);
}


/***********************************************************
    Generate an initial path by forcing the end point.
     x[TTEST] < xMax
************************************************************/
int initPath()
{
    int t;
    int steps=0;
    double  oldX,oldgrX,oldgrY;
    
    x[0]=1.0;
    y[0]=0.0;
        
    for(t=0;t<TBINS;t++) grx[t] =sqrt(D*TSTEP)*genRanGaussian(); /*Generate a gr with correct variance*/
    for(t=0;t<TBINS;t++) gry[t] =sqrt(D*TSTEP)*genRanGaussian(); /*Generate a gr with correct variance*/
        
    propagate(0,TBINS-1);

    oldX =x[TTEST];
    
    /* Force end point of path upto the barrier*/
    while(x[TTEST]>0.0)  
    {
        steps++;    
        t = (int) floor(genRanf()*TTEST);  /* choose gr[] to change*/

        oldgrX = grx[t];
        oldgrY = gry[t];
        grx[t] =sqrt(D*TSTEP)*genRanGaussian(); /*Generate a gr with correct variance*/
        gry[t] =sqrt(D*TSTEP)*genRanGaussian(); /*Generate a gr with correct variance*/
                
        propagate(0,TTEST);
            
        if(oldX<x[TTEST])
        {
            /*REJECT*/
            grx[t]=oldgrX;
            gry[t]=oldgrY;
            
        }
        oldX=x[TTEST];
    }   
    
    return steps;
}

    
/*************************************************
 Propagate path from t=start to t=end
  NOTE: Changes occure from x[start+1] to x[end]
*************************************************/
void propagate(int start, int end)
{
    double x2,xa,ya,xf,yf;
    
    int t;
    
    /*** INSANITY CHECK***/
    if((start<0) || (start>=end) || (end>=TBINS) )
            die("Propagate Error: Start %d \t End:%d",start,end);
                
    
    for(t=start;t<end;t++)
    {
        x2 =x[t]*x[t];
        xa = x[t] + TSTEP*x[t]*(1.-x2-alpha*y[t]*y[t]) + grx[t];
        ya = y[t] - TSTEP*mu*(1.+x2)*y[t] +gry[t];
        
        x2 = xa*xa;
        xf = x[t] + TSTEP*xa*(1.-x2-alpha*ya*ya) + grx[t];
        yf = y[t] - TSTEP*mu*(1.+x2)*ya +gry[t];
        
        x[t+1] =(xf+xa)/2.;
        y[t+1] =(yf+ya)/2.;
    }
}


/*****************************************************
 Propagate path from t=start until condition met,
  or until time runs out.
 Return time of tranition, or TBINS if ran out of time.
 *****************************************************/
int propagateTest(int start)
{
    double x2,xa,ya,xf,yf;
    
    int t;
    
    /*** INSANITY CHECK***/
    if( (start<0) || (start>=TBINS) )
            die("PropagateTest Error: Start %d ",start);
            
    for(t=start;t<(TBINS-1);t++)
    {
        x2 =x[t]*x[t];
        xa = x[t] + TSTEP*x[t]*(1.-x2-alpha*y[t]*y[t]) + grx[t];
        ya = y[t] - TSTEP*mu*(1.+x2)*y[t] +gry[t];
        
        x2 = xa*xa;
        xf = x[t] + TSTEP*xa*(1.-x2-alpha*ya*ya) + grx[t];
        yf = y[t] - TSTEP*mu*(1.+x2)*ya +gry[t];
        
        x[t+1] =(xf+xa)/2.;
        y[t+1] =(yf+ya)/2.;
        
        if(x[t+1]<0.0) return(t+1);
    }
    
    return(TBINS);
}
    


/*************  Misc. Routines **********************************/


/***********************************************************
                Tried and Failed.
tried[] is the number of attempted moves at that time slice.
failed[] is the number of failures.
***********************************************************/
void clearTriedFailed()
{
    int t;
    
    for(t=0;t<TBINS;t++) /*clear tried and failed*/
    {
        failed[t]=0;
        tried[t]=0;
    }
}

void printTriedFailed()
{
    int t;
    
    /*** Print rejection analisis ****/
    
    printf("\nRejection:\t\t\t\t");
    for(t=0;t<TPRINT*TSTEPSPER;t+=SKIP) 
    printf("%lf\t",(double)failed[t]/(double)tried[t]);
    printf("\n");
}



/**************************************
        Print Path
**************************************/
void printTimeAxis()
{
    int t;
    printf("\t\t\t\t");
        for(t= -(TPRINT*TSTEPSPER);t<= +(TPRINT*TSTEPSPER);t+=SKIP)
            printf("%lf\t",(double)t*TSTEP);
    printf("\n");
}
    


void printPaths(int samples, int trans)
{
    int t;
                
    printf("%d\t%d\t%lg\t%lg\t",samples,trans,x[trans],y[trans]);
    for(t=trans-(TPRINT*TSTEPSPER);t<=trans+(TPRINT*TSTEPSPER);t+=SKIP)
    {
        if( (t>0) && (t<(TBINS-1)) ) printf("%lf",x[t]);        
        printf("\t");
    }

    printf("\n\t\t\t\t");
    for(t=trans-(TPRINT*TSTEPSPER);t<=trans+(TPRINT*TSTEPSPER);t+=SKIP)
    {
        if( (t>0) && (t<(TBINS-1)) ) printf("%lf",y[t]);        
        printf("\t");
    }

    
    printf("\n");
    fflush(stdout);
}


void die(char *comment, ...)
{   
    va_list argp;

    va_start(argp, comment);
        vfprintf(stderr, comment, argp);
    va_end(argp);
    fprintf(stderr, "\n");
    
    abort();
}



